<!doctype html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.4/css/all.min.css" rel="stylesheet">
    <link href="http://fonts.cdnfonts.com/css/arial-nova" rel="stylesheet">
    <title>{{ title }}</title>
    {{ resources }}
    {{ bokeh_resources }}
    {{ script }}
</head>
<body>
  <div class="navbar navbar-branded mb-4">
      <div class="container float-left">
        <div>
          <h1>{{ title }}</h1>
          <p class="lead">{{ lead }}</p>
        </div>
        <div class="float-right">
          <span class="float-left sticker-white">RUO</span>
        </div>
      </div>
  </div>
  <div class="container mt-4">
          {{ div }}
</body>
</html>
